import logging
from itertools import chain

from django.conf import settings

from django.utils.encoding import force_text
from django.core.exceptions import ValidationError

from .parsing import Parser
from .errors import InlineValidationError, create_verbose_inline_errors

__all__ = ('Renderer', 'renderer',)


logger = logging.getLogger(__name__)


class Renderer(object):

    def render(self, content, media=None, safe=True,
               raise_errors=False, verbose_errors=True):
        syntax_errors = []
        inline_errors = []
        raise_errors = raise_errors or getattr(settings, 'INLINE_DEBUG', False)

        try:
            nodes, errors = Parser(media=media).parse(content)
            syntax_errors.extend(errors)
        except ValidationError:
            if raise_errors:
                raise

        try:
            content, errors = self.render_nodes(nodes, media, safe)
            inline_errors.extend(errors)
        except ValidationError:
            if raise_errors:
                raise

        if bool(syntax_errors or inline_errors):
            errors = [
                ve[1] for ve in sorted(
                    chain(
                        ((se.lineno, se) for se in syntax_errors),
                        ((ie.lineno, ie) for ie in inline_errors)),
                    key=lambda obj: obj[0])]

            validation_errors = (
                create_verbose_inline_errors(errors) if verbose_errors
                else ValidationError(errors))

            if raise_errors:
                raise validation_errors

        return content

    def render_nodes(self, nodes, media, safe):
        bits = []
        errors = []
        for node in nodes:
            try:
                bits.append(force_text(node.render(media=media, safe=safe)))
            except ValidationError as err:
                errors.extend(
                    (InlineValidationError(node.lineno, msg)
                        for msg in err.messages))
        return u''.join(bits), errors


renderer = Renderer()
